from __future__ import annotations

import logging
from typing import Any, Dict

from actionweaver.actions.orchestration import (
    RequireNext,
    SelectOne,
    _ActionHandlerLLMInvoke,
    _ActionHandlerRequired,
    _ActionHandlerSelectOne,
)
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.pydantic_utils import create_pydantic_model_from_func


class ActionException(Exception):
    pass


def action(
    name,
    scope=DEFAULT_ACTION_SCOPE,
    logger=None,
    orch_expr=None,
    models=[],
    stop=False,
):
    """
    Decorator function to create an Action object.

    Args:
    - name (str): Name of the action.
    - scope (str): Scope of the action, default is DEFAULT_ACTION_SCOPE.
    - logger (logging.Logger): Logger instance to log information, default is None.
    - models (list[pydantic.BaseModel]): List of pydantic models to be used in the action.
    - stop (bool): If True, the agent will stop immediately after invoking this action.
    Returns:
    - create_action: A function that takes a decorated object and returns an Action object.
    """

    _logger = logger or logging.getLogger(__name__)

    def create_action(decorated_obj):
        _logger.debug({"message": f"Creating action with name: {name}, scope: {scope}"})

        action = Action(
            name=name,
            scope=scope,
            decorated_obj=decorated_obj,
            orch_expr=orch_expr,
            logger=_logger,
            stop=stop,
        ).build_pydantic_model_cls(models=models)

        return action

    return create_action


class Action:
    def __init__(
        self,
        name,
        decorated_obj,
        scope=None,
        orch_expr=None,
        logger=None,
        stop=False,
    ):
        self.name = name
        self.scope = scope or DEFAULT_ACTION_SCOPE
        self.logger = logger
        self.orch_expr = orch_expr
        self.stop = stop

        if decorated_obj.__doc__ is None:
            raise ActionException(
                f"Decorated method under action {name} must have a docstring for description."
            )

        self.description = decorated_obj.__doc__
        self.pydantic_cls = None

        self.decorated_method = decorated_obj

    def build_pydantic_model_cls(self, models=None):
        if models is None:
            models = []

        self.pydantic_cls = create_pydantic_model_from_func(
            self.decorated_method, self.decorated_method.__name__.title(), models=models
        )
        return self

    def json_schema(self):
        return self.pydantic_cls.model_json_schema()

    def bind(self, instance) -> InstanceAction:
        return InstanceAction(self, instance)

    def __get__(self, instance, owner) -> InstanceAction:
        """
        Bind a action with an instance.

        Note:
            The `__get__` method is a descriptor method that is called when the action is accessed from an instance.
            It returns an instance-specific action method that is bound to the given instance.
        """
        return InstanceAction(self, instance)


class InstanceAction:
    def __init__(self, action, instance):
        self.action = action
        self.__module__ = action.decorated_method.__module__
        self.__name__ = action.decorated_method.__name__
        self.__qualname__ = action.decorated_method.__qualname__
        self.__annotations__ = action.decorated_method.__annotations__
        self.__doc__ = action.decorated_method.__doc__
        self.instance = instance

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.action.logger:
            self.action.logger.debug(
                {
                    "message": f"[Action {self.action.name}, method {self.__name__}] Calling action: {self.action.name} with args: {args}"
                }
            )
        response = self.action.decorated_method(self.instance, *args, **kwargs)
        if self.action.logger:
            self.action.logger.debug(
                {
                    "message": f"[Action {self.action.name}, method {self.__name__}] Received response: {response}"
                }
            )
        return response


class ActionHandlers:
    def __init__(self, *args, **kwargs):
        self.name_to_action: Dict[str, Action] = {}

    def contains(self, name) -> bool:
        return name in self.name_to_action

    def __len__(self) -> int:
        return len(self.name_to_action)

    def scope(self, scope):
        return {
            name: action
            for name, action in self.name_to_action.items()
            if action.scope == scope
        }

    def bind(self, instance) -> InstanceActionHandlers:
        return InstanceActionHandlers(instance, self)

    def check_orchestration_expr_validity(self, expr):
        if expr is None:
            return

        if isinstance(expr, str):
            if expr not in self.name_to_action:
                raise ActionException(f"Action {expr} not found.")
            return

        for element in expr:
            self.check_orchestration_expr_validity(element)

    @classmethod
    def merge(cls, *handlers) -> ActionHandlers:
        merged = cls()
        for handler in handlers:
            merged.name_to_action.update(handler.name_to_action)
        return merged


class ActionOrchestrationParseError(Exception):
    pass


def parse_orchestration_expr(expr):
    orch_dict = {}

    def get_first_action(l):
        """
        Get first action name from list expression.
        """
        if isinstance(l, (SelectOne, RequireNext)):
            return get_first_action(l[0])
        elif isinstance(l, str):
            return l
        else:
            raise ActionOrchestrationParseError(
                f"Invalid object in orchestration: {l}."
            )

    def get_last_action(l):
        """
        Get last action name from list expression.
        """
        if isinstance(l, SelectOne):
            raise ActionOrchestrationParseError(
                f"Can't decide last action of SelectOne: {l}."
            )
        elif isinstance(l, RequireNext):
            return get_last_action(l[-1])
        elif isinstance(l, str):
            return l
        else:
            raise ActionOrchestrationParseError(
                f"Invalid object in orchestration: {l}."
            )

    def parse(l):
        """
        Parse list expression into orch_dict.
        """

        if isinstance(l, SelectOne):
            curr = get_last_action(l[0])
            parse(l[0])

            next_actions = []
            for n in list(l)[1:]:
                next_actions.append(get_first_action(n))
                parse(n)

            if curr in orch_dict:
                raise ActionOrchestrationParseError(f"Inconsistency caused by {l}")
            orch_dict[curr] = _ActionHandlerSelectOne(next_actions)
        elif isinstance(l, RequireNext):
            parse(l[0])
            prev, curr = get_last_action(l[0]), None
            for n in range(1, len(l)):
                curr = get_first_action(l[n])

                if prev in orch_dict:
                    raise ActionOrchestrationParseError(f"Inconsistency caused by {l}")
                orch_dict[prev] = _ActionHandlerRequired(get_first_action(curr))
                parse(l[n])

                prev = curr
        elif isinstance(l, str):
            return
        else:
            raise ActionOrchestrationParseError(f"Invalid orchestration expr: {l}.")

    parse(expr)
    return orch_dict


class InstanceActionHandlers:
    def __init__(self, instance, action_handlers: ActionHandlers, *args, **kwargs):
        self.action_handlers = action_handlers
        self.instance = instance
        self.orch_dict = {}

    def __getitem__(self, key) -> InstanceAction:
        val = self.action_handlers.name_to_action[key]
        return val.bind(self.instance)

    def __len__(self) -> int:
        return len(self.action_handlers)

    def scope(self, scope):
        return self.action_handlers.scope(scope)

    def contains(self, name) -> bool:
        return self.action_handlers.contains(name)

    def build_orchestration_dict(self):
        """
        Parse orchestration expressions from all actions.
        """
        orch_dict = {
            _ActionHandlerLLMInvoke(DEFAULT_ACTION_SCOPE): _ActionHandlerSelectOne([])
        }

        for _, action in self.action_handlers.name_to_action.items():
            if action.orch_expr:
                if action.orch_expr[0] != action.name:
                    raise ActionOrchestrationParseError(
                        f"Invalid expression {action.orch_expr} in Action {action.name}, "
                        "The first element of the orchestration expression must be the action name. "
                        "For example, in 'action1', 'SelectOne' should be used as 'SelectOne([action1, ...])', "
                        "and 'RequireNext' should be used as 'RequireNext([action1, ...])'."
                    )

                orch_dict.update(**parse_orchestration_expr(action.orch_expr))

            # Parse scope
            if (
                _ActionHandlerLLMInvoke(action.scope) not in orch_dict
                or orch_dict[_ActionHandlerLLMInvoke(action.scope)] is None
            ):  # if scope doesn't exist or map to None like DEFAULT_ACTION_SCOPE, then create a new one
                orch_dict[
                    _ActionHandlerLLMInvoke(action.scope)
                ] = _ActionHandlerSelectOne([])
            elif not isinstance(
                orch_dict[_ActionHandlerLLMInvoke(action.scope)],
                _ActionHandlerSelectOne,
            ):
                raise ActionOrchestrationParseError(
                    f"The scope {action.scope} of Action {action.name} causes inconsistency in orchestration."
                )
            orch_dict[_ActionHandlerLLMInvoke(action.scope)].append(action.name)

        self.orch_dict = orch_dict
        return self

    def to_pyvis_network(self):
        """Experimental for debugging purpose"""
        from pyvis.network import Network

        def _generate_node_config(n):
            if isinstance(n, _ActionHandlerLLMInvoke):
                if n.scope == DEFAULT_ACTION_SCOPE:
                    return {
                        "n_id": str(n),
                        "label": "LLM[DEFAULT]",
                        "color": "#005B00",
                        "physics": False,
                    }
                else:
                    return {
                        "n_id": str(n),
                        "label": f"LLM[{n.scope}]",
                        "color": "#00CC99",
                        "physics": False,
                    }
            elif isinstance(n, str):
                return {
                    "n_id": n,
                    "label": None,
                    "color": "#FF7769",
                    "physics": False,
                }
            else:
                raise ActionException("unsupported node type for visualization")

        g = Network(notebook=True, directed=True)

        for node, expr in self.orch_dict.items():
            g.add_node(**_generate_node_config(node))

            if isinstance(expr, _ActionHandlerSelectOne):
                for dest in expr:
                    g.add_node(**_generate_node_config(dest))
                    g.add_edge(
                        _generate_node_config(node)["n_id"],
                        _generate_node_config(dest)["n_id"],
                        color="#00CC99",
                        label="select",
                        physics=False,
                    ),
            elif isinstance(expr, _ActionHandlerRequired):
                g.add_node(**_generate_node_config(expr.action))
                g.add_edge(
                    _generate_node_config(node)["n_id"],
                    _generate_node_config(expr.action)["n_id"],
                    color="#FF0000",
                    label="requires",
                    physics=False,
                ),
        return g



import json
import logging
import time
import uuid

import openai
from openai.openai_object import OpenAIObject

from actionweaver.actions.action import (
    ActionHandlers,
    InstanceActionHandlers,
    parse_orchestration_expr,
)
from actionweaver.actions.orchestration import _ActionDefault, _ActionHandlerLLMInvoke
from actionweaver.llms.openai.functions import Functions
from actionweaver.llms.openai.tokens import TokenUsageTracker
from actionweaver.utils import DEFAULT_ACTION_SCOPE
from actionweaver.utils.stream import (
    get_first_element_and_iterator,
    is_generator,
    merge_dicts,
)


class OpenAIChatCompletionException(Exception):
    pass


class OpenAIChatCompletion:
    def __init__(self, model, token_usage_tracker=None, logger=None):
        self.model = model
        self.instance_action_handlers = InstanceActionHandlers(
            None, ActionHandlers()
        ).build_orchestration_dict()
        self.logger = logger or logging.getLogger(__name__)
        self.token_usage_tracker = token_usage_tracker or TokenUsageTracker(
            logger=logger
        )

    def _bind_action_handlers(self, instance_action_handlers: InstanceActionHandlers):
        self.instance_action_handlers = instance_action_handlers

    def _invoke_function(
        self,
        call_id,
        messages,
        model,
        function_call,
        orchestration_dict,
        default_expr,
        functions,
    ):
        """Invoke the function, update the messages, returns functions argument for the next OpenAI API call or halt the function loop and return the response."""

        if isinstance(function_call, OpenAIObject):
            function_call = function_call.to_dict()

        messages += [
            {
                "role": "assistant",
                "content": None,
                "function_call": function_call,
            }
        ]

        name = function_call["name"]
        arguments = function_call["arguments"]

        if self.instance_action_handlers.contains(name):
            try:
                arguments = json.loads(function_call["arguments"])
            except json.decoder.JSONDecodeError as e:
                self.logger.error(
                    {
                        "message": "Parsing function call arguments from OpenAI response ",
                        "arguments": function_call["arguments"],
                        "timestamp": time.time(),
                        "model": model,
                        "call_id": call_id,
                    },
                    exc_info=True,
                )
                raise OpenAIChatCompletionException(e) from e

            # Invoke action
            function_response = self.instance_action_handlers[name](**arguments)
            stop = self.instance_action_handlers[name].action.stop
            messages += [
                {
                    "role": "function",
                    "name": name,
                    "content": str(function_response),
                }
            ]

            self.logger.debug(
                {
                    "message": "Action invoked and response received",
                    "action_name": name,
                    "action_arguments": arguments,
                    "action_response": function_response,
                    "timestamp": time.time(),
                    "call_id": call_id,
                    "stop": stop,
                }
            )

            # Update new functions for next OpenAI api call
            # if name doesn't exist in orchestration dict, use _ActionDefaultLLM which doesn't invoke functions
            expr = (
                orchestration_dict[name]
                if name in orchestration_dict
                else _ActionDefault()
            )
            return (
                Functions.from_expr(
                    expr,
                    self.instance_action_handlers,
                ),
                (stop, function_response),
            )
        else:
            unavailable_function_msg = f"{name} is not a valid function name, use one of the following: {', '.join([func['name'] for func in functions.functions])}"
            messages += [
                {
                    "role": "user",
                    "content": unavailable_function_msg,
                }
            ]
            self.logger.debug(
                {
                    "message": "Unavailable action",
                    "action_name": name,
                    "action_arguments": arguments,
                    "action_response": unavailable_function_msg,
                    "timestamp": time.time(),
                    "call_id": call_id,
                }
            )
            return functions, (False, None)

    def create(
        self, messages, *args, scope=None, orch_expr=None, stream=False, **kwargs
    ):
        """
        Invoke the OpenAI API with the provided messages and functions.

        Functions supplied to the API are determined by the specified scope and orchestration expression.

        Args:
            messages (list): List of message objects for the conversation.
            scope (str): Scope of the functions to be used.
            orch_expr: If specified, overrides the orchestration dictionary, determining the API call flow.
            stream (bool): If True, returns a generator that yields the API responses.

        Returns:
            API response with generated output.
        """

        # Todo: pass call_id to the decorated method
        call_id = str(uuid.uuid4())

        model = kwargs.get("model", self.model)
        temperature = kwargs.get("temperature", 0.0)

        # Restart token usage tracker
        self.token_usage_tracker.clear()

        if scope is None:
            scope = DEFAULT_ACTION_SCOPE
        default_expr = _ActionHandlerLLMInvoke(scope)

        orchestration_dict = self.instance_action_handlers.orch_dict
        # If an orchestration expression is provided, override the orchestration dictionary
        if orch_expr is not None:
            self.instance_action_handlers.action_handlers.check_orchestration_expr_validity(
                orch_expr
            )

            scope = "_llm_orchestration"
            # Construct a new orchestration dictionary using the parsed orchestration expression
            new_orchestration_dict = parse_orchestration_expr([scope] + orch_expr)

            # Replace the default expression with the new one
            default_expr = _ActionHandlerLLMInvoke(scope)
            new_orchestration_dict[default_expr] = new_orchestration_dict.pop(scope)

            # Update the orchestration dictionary with the new one
            orchestration_dict = new_orchestration_dict

        self.logger.debug(
            {
                "message": "Creating new chat completion",
                "model": model,
                "scope": scope,
                "input_messages": messages,
                "timestamp": time.time(),
                "call_id": call_id,
            }
        )

        response = None
        functions = Functions.from_expr(
            orchestration_dict[default_expr],
            self.instance_action_handlers,
        )

        while True:
            self.logger.debug(
                {
                    "message": "Calling OpenAI API",
                    "call_id": call_id,
                    "input_messages": messages,
                    "model": model,
                    "scope": scope,
                    "timestamp": time.time(),
                    **functions.to_arguments(),
                }
            )

            function_argument = functions.to_arguments()
            if function_argument["functions"]:
                api_response = openai.ChatCompletion.create(
                    model=model,
                    temperature=temperature,
                    messages=messages,
                    stream=stream,
                    **function_argument,
                )
            else:
                api_response = openai.ChatCompletion.create(
                    model=model,
                    temperature=temperature,
                    messages=messages,
                    stream=stream,
                )

            # logic to handle streaming API response
            if is_generator(api_response):
                first_element, iterator = get_first_element_and_iterator(api_response)

                if "content" in first_element.choices[0]["delta"] and isinstance(
                    first_element.choices[0]["delta"]["content"], str
                ):
                    # if the first element is a message, return generator right away.
                    return iterator
                elif "function_call" in first_element.choices[0]["delta"]:
                    # if the first element in generator is a function call, merge all the deltas.
                    l = list(iterator)

                    deltas = {}
                    for element in l:
                        delta = element["choices"][0]["delta"].to_dict()
                        deltas = merge_dicts(deltas, delta)

                    first_element["choices"][0]["message"] = deltas
                    first_element["choices"][0]["delta"] = deltas
                    api_response = first_element
                else:
                    raise OpenAIChatCompletionException(
                        f"Unsupported response from streaming API: {list(iterator)}"
                    )

            else:
                self.token_usage_tracker.track_usage(api_response.usage)

            self.logger.debug(
                {
                    "message": "Received response from OpenAI API",
                    "response": api_response,
                    "timestamp": time.time(),
                    "call_id": call_id,
                }
            )

            choice = api_response.choices[0]
            message = choice["message"]

            if "function_call" in message and message["function_call"]:
                functions, (stop, resp) = self._invoke_function(
                    call_id,
                    messages,
                    model,
                    message["function_call"],
                    orchestration_dict,
                    default_expr,
                    functions,
                )
                if stop:
                    return resp
            elif "content" in message and message["content"]:
                response = message["content"]

                # ignore last message in the function loop
                # messages += [{"role": "assistant", "content": message["content"]}]
                if choice["finish_reason"] == "stop":
                    """
                    Stop Reasons:

                    - Occurs when the API returns a message that is complete or is concluded by one of the stop sequences defined via the 'stop' parameter.

                    See https://platform.openai.com/docs/guides/gpt/chat-completions-api for details.
                    """
                    self.logger.debug(
                        {
                            "message": "Model decides to stop",
                            "model": model,
                            "timestamp": time.time(),
                            "call_id": call_id,
                        }
                    )

                    break
            else:
                raise OpenAIChatCompletionException(
                    f"Unsupported response from OpenAI api: {api_response}"
                )
        return response



from __future__ import annotations

from tkinter import N

from actionweaver.actions.action import Action, ActionHandlers
from actionweaver.llms.openai import chat
from actionweaver.llms.openai.chat import OpenAIChatCompletion


class ActionHandlerMixinException(Exception):
    pass


class ActionHandlerMixin:
    _action_handlers = ActionHandlers()

    def __post_init__(self):
        # check if all action orchestration expressions are valid
        for _, action in self._action_handlers.name_to_action.items():
            if action.orch_expr:
                if len(action.orch_expr) < 2:
                    raise ActionHandlerMixinException(
                        f"Action {action.name} must has at least two elements in its orchestration expression. The first element is the action itself. For example, SelectOne([{action.name}, action1, action2])."
                    )
                self._action_handlers.check_orchestration_expr_validity(
                    action.orch_expr
                )

        # bind action handlers to self
        self.instance_action_handlers = self._action_handlers.bind(self)

        # build action orchestration dict
        self.instance_action_handlers.build_orchestration_dict()

        chat_completion_found = False
        for _, attr_value in tuple(self.__dict__.items()):
            if isinstance(attr_value, OpenAIChatCompletion):
                if chat_completion_found:
                    raise ActionHandlerMixinException(
                        "Only one OpenAIChatCompletion instance is allowed in a class."
                    )

                # bind instance action handlers to llm
                attr_value._bind_action_handlers(self.instance_action_handlers)
                chat_completion_found = True

        if not chat_completion_found:
            raise ActionHandlerMixinException(
                "An OpenAIChatCompletion instance is required in a class."
            )

    @classmethod
    def __post_init_subclass__(cls, **kwargs):
        # add action to action handlers.
        for _, attr_value in tuple(cls.__dict__.items()):
            if isinstance(attr_value, Action):
                cls._action_handlers.name_to_action[attr_value.name] = attr_value

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        def init_decorator(original_init):
            def new_init(self, *args, **kwargs):
                try:
                    original_init(self, *args, **kwargs)
                except Exception as e:
                    error_message = (
                        f"Error occurred while initializing the object: {e}."
                    )
                    raise ActionHandlerMixinException(error_message) from e

                if type(self) == cls:
                    self.__post_init__()

            return new_init

        def check_orchestration_expr_validity(expr):
            if expr is None:
                return

            if isinstance(expr, str):
                if expr not in cls._action_handlers.name_to_action:
                    raise ActionHandlerMixinException(
                        f"Action {expr} not found in {cls.__name__}."
                    )
                return

            for element in expr:
                check_orchestration_expr_validity(element)

        # Add __post_init__ method to the subclass.
        cls.__init__ = init_decorator(cls.__init__)

        # create action handlers from base class.
        cls._action_handlers = ActionHandlers.merge(
            *[
                base_cls._action_handlers
                for base_cls in cls.__bases__
                if hasattr(base_cls, "_action_handlers")
            ]
        )

        cls.__post_init_subclass__(kwargs=kwargs)

    def action_to_pyvis_network(self):
        if self.instance_action_handlers is None:
            raise ActionHandlerMixinException(
                "Action handlers not initialized. Please call __post_init__ first."
            )

        return self.instance_action_handlers.to_pyvis_network()
